@@ -1,5 +1,5 @@
 /*
- *      Copyright (C) 1997-1999 Claus-Justus Heine
+ *      Copyright (C) 1997-2000 Claus-Justus Heine
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
@@ -22,11 +22,11 @@
  */
 
 char src[] = "$RCSfile: libftvt.c,v $";
-char rev[] = "$Revision: 1.10 $";
-char dat[] = "$Date: 2000/06/25 16:05:47 $";
+char rev[] = "$Revision: 1.14 $";
+char dat[] = "$Date: 2000/07/22 18:53:01 $";
 
 #ifdef HAVE_CONFIG_H
-#include <config.h>
+# include <config.h>
 #endif
 
 #include <sys/types.h>
@@ -43,6 +43,8 @@ char dat[] = "$Date: 2000/06/25 16:05:47 $";
 #include <ctype.h>
 #include <stdarg.h>
 #include <time.h>
+#include <iconv.h>
+#include <wchar.h>
 #include <libintl.h>
 #define _(String) gettext (String)
 
@@ -54,6 +56,14 @@ char dat[] = "$Date: 2000/06/25 16:05:47 $";
 #include "ftvt.h"
 #include "libftvt-version.h"
 
+#if !HAVE_MINOR
+# ifdef minor
+#  undef minor
+# endif
+/* glibc-2.0 seems not to include the "minor()" macro by default */
+# define minor(dev) ((int)((dev) & 0xff))
+#endif
+
 extern void *xmalloc(size_t size);
 extern void *xrealloc(void *old, size_t size);
 
@@ -75,9 +85,6 @@ static const char *drivetype;
 static void ftvt_error(const char *fmt, ...)
 	__attribute__ ((format (printf, 1, 2)));
 static void ftvt_perror(const char *msg);
-static int ftvt_read_header_segment(int tape_fd,
-									int *first_data_segment,
-									int *last_data_segment);
 static char *ftvt_gets(FILE *fp, char *buffer, int *len);
 
 
@@ -196,7 +203,7 @@ static int raw_mode_onoff(int tape_fd, int onoff)
 	mtftmode.ft_rawmode = 0;
 	if (ioctl(tape_fd, MTIOCFTMODE, &mtftmode) == -1) {
 		/* not supported */
-#ifdef ZFT_OBSOLETE
+#ifdef CONFIG_ZFT_OBSOLETE
 		/* Ok, in this case we don't really switch to and from raw
 		 * mode, but simply check for the correct device.
 		 */
@@ -343,20 +350,59 @@ int ftvt_close(int tape_fd)
  */
 int ftvt_read(int tape_fd, ftvt *volumes, u_int8_t *buffer)
 {
-	const struct mtop rewind = { MTREW, 1 };
+	u_int8_t hseg[FT_SEGMENT_SIZE];
+
+	if ((fmt_code = ftvt_read_header_segment(tape_fd, hseg)) == -1) {
+		return -1;
+	}
+
+	return ftvt_read_vtbl(tape_fd, hseg, fmt_code, volumes, buffer);
+}
+
+/* Read the volume table. This function requires the header segment
+ * of the tape cartridge to determine the location of the volume table
+ * segment, and then reads volume table segment.
+ *
+ * Arguments:
+ *
+ * int tape_fd  : The file descriptor previously return by ftvt_open()
+ *
+ * u_int8_t *hseg:Buffer which should contain an image of the header segment.
+ *
+ * fmt_code     : Format code of the tape, returned by ftvt_read_header_segment.
+ *
+ * ftvt *volumes: Buffer to hold the volume table in. The size of the
+ *                buffer must be FTVT_MAX_VOLUMES (i.e. 29*1024/128)
+ *
+ * u_int8_t     : The buffer to hold the contents of the volume table
+ *                segment. It must be of size 29*1024.
+ *
+ * Return value : The number of volumes found in the volume table segment.
+ *
+ * Side effects : The internal variable "max_volumes" is set to the
+ *                number of volumes which would fit into the volume
+ *                table segment. Normally this should be FTVT_MAX_VOLUMES,
+ *                but some cartridges use volume table segments which
+ *                contain bad sectors.
+ */
+int ftvt_read_vtbl(int tape_fd, u_int8_t *hseg, int fmt_code, ftvt *volumes,
+				   u_int8_t *buffer)
+{
 	struct mtftseg ft_seg;
 	int vtbl_cnt; 
 	int end_seg = 0;
 	const char *ids[] = FTVT_IDS;
 
-	if (ioctl(tape_fd, MTIOCTOP, &rewind) == -1) {
-		ftvt_perror(_("Ioctl error rewinding tape"));
-		return -1;
-	}
-	if ((fmt_code = ftvt_read_header_segment(tape_fd,
-											 &first_seg, &last_seg)) == -1) {
-		return -1;
+	if (fmt_code == fmt_big) {
+		memcpy(&first_seg, &hseg[FT_6_FRST_SEG], 4);
+		memcpy(&last_seg,  &hseg[FT_6_LAST_SEG], 4);
+	} else {
+		memcpy(&first_seg, &hseg[FT_FRST_SEG], 2);
+		memcpy(&last_seg,  &hseg[FT_LAST_SEG], 2);
+		first_seg &= 0x0000ffff;
+		last_seg &= 0x0000ffff;
 	}
+
 	if (verbose) {
 		printf(_("Reading volume table segment ... "));
 		fflush(stdout);
@@ -396,16 +442,35 @@ int ftvt_read(int tape_fd, ftvt *volumes, u_int8_t *buffer)
 		   memcmp(buffer, ids[3], 4) == 0) {
 
 		memcpy(&volumes[vtbl_cnt].entry, buffer, FTVT_SIZE);
+#if 0
+		memcpy(&volumes[vtbl_cnt].bytes, buffer, FTVT_SIZE);
+#else
+		volumes[vtbl_cnt].bytes = (u_int8_t *)&volumes[vtbl_cnt].entry;
+#endif
 		if (memcmp(&volumes[vtbl_cnt].ftvt_sig, "VTBL", 4) == 0) {
 			volumes[vtbl_cnt].fmt_code = fmt_code;
 			if (fmt_code == fmt_big) {
-				volumes[vtbl_cnt].start = end_seg + 1;
-				volumes[vtbl_cnt].end   = (end_seg + 
-										   volumes[vtbl_cnt].ftvt_space);
+				/* work-around for ftape-4.03 bug */
+				if (memcmp(volumes[vtbl_cnt].entry.ext.ext, ZFT_SIG,
+						   ZFT_SIGLEN) == 0) {
+					volumes[vtbl_cnt].zft_bug = 1;
+					volumes[vtbl_cnt].start   = end_seg + 1;
+					volumes[vtbl_cnt].end     = (end_seg + 
+												 volumes[vtbl_cnt].ftvt_space);
+				} else {
+					volumes[vtbl_cnt].zft_bug = 0;
+					volumes[vtbl_cnt].start   = 
+						*(u_int32_t *)volumes[vtbl_cnt].ftvt_6_start;
+					volumes[vtbl_cnt].end     =
+						*(u_int32_t *)volumes[vtbl_cnt].ftvt_6_end;
+				}
+				volumes[vtbl_cnt].num_segments = volumes[vtbl_cnt].ftvt_space;
 				end_seg = volumes[vtbl_cnt].end;
 			} else {
 				volumes[vtbl_cnt].start = volumes[vtbl_cnt].ftvt_start;
 				volumes[vtbl_cnt].end   = volumes[vtbl_cnt].ftvt_end;
+				volumes[vtbl_cnt].num_segments = 
+					volumes[vtbl_cnt].end - volumes[vtbl_cnt].start + 1;
 			}
 			volumes[vtbl_cnt].fmt_code = fmt_code;
 		}
@@ -417,16 +482,24 @@ int ftvt_read(int tape_fd, ftvt *volumes, u_int8_t *buffer)
 }
 
 /* Try to read the header segments. Return the format code of this
- * cartridge, and set *first_data_segment and *last_data_segment.
+ * cartridge.
+ *
+ * Arguments:
+ *
+ * int tape_fd   : The file descriptor previously return by ftvt_open()
+ *
+ * u_int8_t *hseg: The buffer to hold the contents of the header
+ *                 segment. It must be of size 29*1024 (FT_SEGMENT_SIZE).
+ *
+ * Return value  : The format code of the tape.
+ *
+ * Side effects  : None
  */
-static int ftvt_read_header_segment(int tape_fd,
-									int *first_data_segment,
-									int *last_data_segment)
+int ftvt_read_header_segment(int tape_fd, u_int8_t *hseg)
 {
 	struct mtftseg ft_seg;
 	int i;
 	const unsigned long hseg_magic = FT_HSEG_MAGIC;
-	u_int8_t hseg[FT_SEGMENT_SIZE];
 	
 	if (verbose) {
 		printf(_("Reading header segment ... "));
@@ -452,15 +525,6 @@ static int ftvt_read_header_segment(int tape_fd,
 	if (verbose) {
 		printf(_("done.\n"));
 	}
-	if (hseg[FT_FMT_CODE] == fmt_big) {
-		memcpy(first_data_segment, &hseg[FT_6_FRST_SEG], 4);
-		memcpy(last_data_segment,  &hseg[FT_6_LAST_SEG], 4);
-	} else {
-		memcpy(first_data_segment, &hseg[FT_FRST_SEG], 2);
-		memcpy(last_data_segment,  &hseg[FT_LAST_SEG], 2);
-		*first_data_segment &= 0x0000ffff;
-		*last_data_segment &= 0x0000ffff;
-	}
 	return (int)hseg[FT_FMT_CODE];
 }
 
@@ -528,29 +592,71 @@ void ftvt_print_one(const ftvt *volume)
 	}
 }
 
-/* Print one volume table in tagged output format (i.e. keyword-value
- * pairs)
- *
- * Arguments:
- *
- * const ftvt *volume: pointer to the volume table entry to print
- */
-void ftvt_print_one_tagged(const ftvt *volume)
+static void print_a_char(FILE *where,
+						 const char *prefix, char c, const char *suffix)
+{
+	if (isprint(c)) {
+		fprintf(where, "%s%c%s", prefix, c, suffix);
+	} else {
+		fprintf(where, "%s\\%03o%s", prefix, (unsigned int)c, suffix);
+	}
+}
+
+static void print_hex_line(FILE *where,
+						   const char *prefix,
+						   const char *what,
+						   const char *suffix,
+						   int howmuch)
+{
+	int i;
+
+	fprintf(where, "%s", prefix);
+	for (i = 0; i < howmuch-1; i++) {
+			printf("0x%02x ", what[i]);
+	}
+	printf("0x%02x%s", what[i], suffix);
+}
+
+static void print_char_line(FILE *where,
+							const char *prefix,
+							const char *what,
+							const char *suffix,
+							int howmuch)
+{
+	int i;
+
+	fprintf(where, "%s", prefix);
+	for (i = 0; i < howmuch-1; i++) {
+		print_a_char(where, "", what[i], " ");
+	}
+	print_a_char(where, "", what[i], "");
+	fprintf(where, "%s", suffix);
+}
+
+/* print a volume table entry with signature "VTBL" */
+static void print_one_vtbl(const ftvt *volume)
 {
 	char sig[5];
 	char label[45];
 	int i;
+	const char *formats[] = { "Extended Format",
+							  "DOS Basic Format",
+							  "Unix",
+							  "OS/2",
+							  "Novell NetWare",
+							  "Windows NT",
+							  "DOS Extended Format",
+							  "Windows 95" };
 
-	memcpy(sig, volume->ftvt_sig, 4);
-	sig[4] = '\0';
-	memcpy(label, volume->ftvt_desc, 44);
-	label[44] = '\0';
-	printf("SIGNATURE \"%s\"\n", sig);
-	if (strcmp(sig, "VTBL") == 0) {
-		printf("START %d\n", volume->start);
-		printf("END %d\n", volume->end);
-		printf("DESCRIPTION \"%s\"\n", label);
-		printf("DATE \"%s\"\n", ftvt_decode_date(volume->ftvt_date));
+	printf("START %d\n", volume->start);
+	printf("END %d\n", volume->end);
+	printf("NUM_SEGMENTS %d\n", volume->num_segments);
+	printf("DESCRIPTION \"%s\"\n", label);
+	printf("DATE \"%s\"\n", ftvt_decode_date(volume->ftvt_date));
+	if (!volume->entry.vendor_specific ||
+		(volume->entry.vendor_specific &&
+		 volume->entry.ext.qic113.major == 113 &&
+		 volume->fmt_code != fmt_big)) {
 		printf("FLAG_VENDOR_SPECIFIC %d\n", volume->entry.vendor_specific);
 		printf("FLAG_MULTI_CARTRIDGE %d\n", volume->entry.multi_cartridge);
 		printf("FLAG_NOT_VERIFIED %d\n", volume->entry.not_verified);
@@ -561,16 +667,35 @@ void ftvt_print_one_tagged(const ftvt *volume)
 		printf("FLAG_RESERVED_6 %d\n", volume->entry.fl_reserved_6);
 		printf("FLAG_RESERVED_7 %d\n", volume->entry.fl_reserved_7);
 		printf("MULTI_CARTRIDGE_COUNT %d\n", volume->entry.m_no);
-		printf("VENDOR_EXTENSION \"");
-		for (i = 0; i < sizeof(volume->entry.ext)-1; i ++) {
-			printf("0x%02x ", volume->entry.ext[i]);
+		if (volume->entry.ext.qic113.major == 113) {
+			printf("QIC_MAJOR %d\n", volume->entry.ext.qic113.major);
+			printf("QIC_MINOR %d\n", volume->entry.ext.qic113.minor);
+			print_hex_line(stdout,
+						   "VENDOR_EXTENSION \"",
+						   volume->entry.ext.qic113.ext,
+						   "\"\n", sizeof(volume->entry.ext.qic113.ext));
+			print_char_line(stdout,
+							"# ",
+							volume->entry.ext.qic113.ext,
+							"\n", sizeof(volume->entry.ext.qic113.ext));
+		} else {
+			print_hex_line(stdout,
+						   "VENDOR_EXTENSION \"",
+						   volume->entry.ext.ext,
+						   "\"\n", sizeof(volume->entry.ext.ext));
+			print_char_line(stdout,
+							"# ",
+							volume->entry.ext.ext,
+							"\n", sizeof(volume->entry.ext.ext));
 		}
-		printf("0x%02x\"\n", volume->entry.ext[i]);
-		printf("PASSWORD \"");
-		for (i = 0; i < sizeof(volume->entry.pwd)-1; i ++) {
-			printf("0x%02x ", volume->entry.pwd[i]);
+		print_hex_line(stdout, "PASSWORD \"", volume->entry.pwd, "\"\n",
+					   sizeof(volume->entry.pwd));
+		if (volume->entry.pwd[0] != '\0') {
+			print_char_line(stdout, "# ", volume->entry.pwd, "\n",
+							sizeof(volume->entry.pwd));
+		} else {
+			printf("# no password\n");
 		}
-		printf("0x%02x\"\n", volume->entry.pwd[i]);
 		printf("DIRECTORY_SIZE %d\n", volume->entry.dir_size);
 		printf("DATA_SIZE %Ld\n", volume->entry.data_size);
 		printf("OS_VERSION 0x%04x\n", volume->entry.os_version);
@@ -579,8 +704,167 @@ void ftvt_print_one_tagged(const ftvt *volume)
 		printf("RESERVED_1 0x%02x\n", volume->entry.reserved_1);
 		printf("COMPRESSION_FLAGS 0x%02x\n", volume->entry.cmpr);
 		printf("FORMAT 0x%02x\n", volume->entry.format);
+		if (volume->entry.format < NR_ITEMS(formats)) {
+			printf("# %s\n", formats[volume->entry.format]);
+		} else {
+			printf("# unknown format and OS Type\n");
+		}
 		printf("RESERVED_2 0x%02x\n", volume->entry.reserved_1);
 		printf("RESERVED_3 0x%02x\n", volume->entry.reserved_1);
+	} else {
+		/* print remaining values as hex string */
+		print_hex_line(stdout,
+					   "VENDOR_DATA \"", &volume->bytes[56], "\"\n", 72);
+		print_char_line(stdout,
+						"# ", &volume->bytes[56], "\n", 72);
+	}
+}
+
+#if 0 /* to be implemented */
+/* convert between character sets, return amount of data stored in
+ * outbuf or -1 on error. Stolen from info manual of GNU libc (iconv
+ * example).
+ */
+static size_t char_conv(const char *to, char *outbuf, size_t outsize,
+						const char *from, const char *inbuf, size_t insize)
+{
+	char *wrptr = outbuf;
+	const char *rdptr = inbuf;
+	iconv_t cd;
+	size_t nconv;
+	
+	cd = iconv_open (to, from);
+	if (cd == (iconv_t) -1) {
+		/* Something went wrong. */
+		if (errno == EINVAL) {
+			ftvt_error("conversion from `%s' to `%s' no available", from, to);
+		} else {
+			ftvt_perror("iconv_open");
+		}
+    
+		/* Terminate the output string.  */
+		*outbuf = '\0';
+		return (size_t)-1;
+	}
+
+	nconv = iconv (cd, (const char **)&rdptr, &insize, &wrptr, &outsize);
+	if (nconv == (size_t) -1) {
+		/* Not everything went right. It might only be
+	     * an unfinished byte sequence at the end of the
+		 * buffer. Or it is a real problem.
+		 */
+		if (errno != EINVAL) {
+			ftvt_perror("iconv");
+		}
+		return (size_t)-1;
+	}
+	     
+	if (iconv_close (cd) != 0) {
+		ftvt_perror("iconv_close");
+	}
+     
+	return (size_t)(wrptr - outbuf);
+}
+
+static void print_ucs2(FILE *where,
+					   const char *prefix,
+					   const char *what,
+					   const char *suffix,
+					   int howmuch)
+{
+	wchar_t *ucs4 = (wchar_t *)alloca(howmuch*2);
+	char *str = (char *)alloca(howmuch/2+1);
+	int nconv;
+
+	nconv = char_conv("UCS4", ucs4, howmuch*2, "UCS2", what, howmuch);
+
+	if (nconv > 0 && nconv < howmuch) {
+		
+	}
+}
+#endif
+
+/* print a volume table entry with signature "XTBL". The entry
+ * contains the volume name and password in unicode UCS2 format
+ * (i.e. 16 bit characters).
+ *
+ * We print the unicode entry as hex-byte string in the form "0x01
+ * 0x02". We do NOT use 4 digit hex notation to not have to care about
+ * byte ordering. Two of our 2 digit hex numbers make one unicode
+ * character, which probably is in big endian order.
+ *
+ * TODO:
+ *
+ * We print a representation of the uni code field using the current's
+ * locale charset as comment.
+ *
+ * FIXME: how to tell iconv to convert to the current locale's
+ * charset? Or how to determine the standard name of the charset of
+ * the current locale? :(
+ */
+static void print_one_xtbl(const ftvt *volume)
+{
+	print_hex_line(stdout,
+				   "NAME \"", &volume->bytes[XTBL_NAME], "\"\n", 88);
+	print_hex_line(stdout, 
+				   "PASSWORD \"", &volume->bytes[XTBL_PWD], "\"\n", 16);
+	print_hex_line(stdout,
+				   "RESERVED \"", &volume->bytes[XTBL_RES], "\"\n", 20);
+}
+
+static void print_one_utid(const ftvt *volume)
+{
+	print_hex_line(stdout,
+				   "NAME \"", &volume->bytes[UTID_NAME], "\"\n", 88);
+	print_hex_line(stdout,
+				   "RESERVED \"", &volume->bytes[XTBL_RES], "\"\n", 36);
+}
+
+static void print_one_exvt(const ftvt *volume)
+{
+	printf("PARENT %d\n", *(u_int16_t *)&volume->bytes[EXVT_PARENT]);
+	printf("CHILD %d\n", *(u_int16_t *)&volume->bytes[EXVT_CHILD]);
+	print_hex_line(stdout,
+				   "RESERVED \"", &volume->bytes[EXVT_RES], "\"\n", 120);
+}
+
+/* Print one volume table in tagged output format (i.e. keyword-value
+ * pairs)
+ *
+ * Arguments:
+ *
+ * const ftvt *volume: pointer to the volume table entry to print
+ */
+void ftvt_print_one_tagged(const ftvt *volume)
+{
+	struct {
+		const char *sig;
+		void (*prt_fct)(const ftvt *volume);
+	} prt_handler[] = {
+		{ "VTBL", print_one_vtbl },
+		{ "XTBL", print_one_xtbl },
+		{ "UTID", print_one_utid },
+		{ "EXVT", print_one_exvt }
+	};
+	char sig[5];
+	char label[45];
+	int i;
+
+	memcpy(sig, volume->ftvt_sig, 4);
+	sig[4] = '\0';
+	memcpy(label, volume->ftvt_desc, 44);
+	label[44] = '\0';
+	printf("SIGNATURE \"%s\"\n", sig);
+	for (i = 0; i < NR_ITEMS(prt_handler); i++) {
+		if (memcmp(sig, prt_handler[i].sig, 4) == 0) {
+			prt_handler[i].prt_fct(volume);
+			break;
+		}
+	}
+	if (i == NR_ITEMS(prt_handler)) { /* unknown signature */
+		print_hex_line(stdout,
+					   "ENTRY_DATA \"", &volume->bytes[4], "\"\n", 124);
+		print_char_line(stdout, "# ", &volume->bytes[4], "\n", 124);
 	}
 }
 
@@ -666,9 +950,14 @@ static char *ftvt_gets(FILE *fp, char *buffer, int *len)
 		}												\
 		if (strrchr(value, '\"')) {						\
 			*strrchr(value, '\"') = '\0';				\
+														\
+		}												\
+		if (dest) {										\
+			memset((dest), 0, max);						\
+			strncpy((dest), value, max);				\
 		}												\
-		if (dest) memcpy((dest), value, max);			\
 		action;											\
+		non_signature_seen = 1;							\
 		continue;										\
 	}													\
 }
@@ -691,11 +980,14 @@ int ftvt_parse_tagged(ftvt *volumes)
 	char *new;
 	int len = 1024;
 	int in_volume = 0;
+	int non_signature_seen = 0;
+	enum { vtbl = 0, xtbl = 1, utid = 2, exvt = 3 } vtbl_id;
 	int num;
 	ftvt *volume = NULL;
 	int get_data = 0;
 	int maxnum = 0;
 	char dummy[6];
+	const char *ftvt_ids[] = FTVT_IDS;
 
 	while ((new = ftvt_gets(stdin, buffer, &len))) {
 		buffer = new;
@@ -737,6 +1029,8 @@ int ftvt_parse_tagged(ftvt *volumes)
 					return -1;
 				}
 				in_volume = 1;
+				non_signature_seen = 0;
+				vtbl_id = vtbl;
 				volume = &volumes[num];
 				volume->modified = 1;
 				volume->num = num;
@@ -751,87 +1045,216 @@ int ftvt_parse_tagged(ftvt *volumes)
 			ftvt_error(_("Corrupt volume input data\n"));
 			return -1;
 		}
-		if (strncmp("SIGNATURE", buffer, sizeof("SIGNATURE")-1) == 0) {
-			if (!strstr(buffer, "VTBL")) {
+		GET_STRING_VALUE(SIGNATURE, volume->ftvt_sig, 4,
+		{
+			int i;
+
+			if (non_signature_seen) {
+					ftvt_error(_("SIGNATURE tag must follow ENTRY at once.\n"));
+					return -1;
+			}
+			for (i = 0; i < NR_ITEMS(ftvt_ids); i++) {
+				if (memcmp(volume->ftvt_sig, ftvt_ids[i], 4) == 0) {
+					break;
+				}
+			}
+			if (i == NR_ITEMS(ftvt_ids)) {
 				ftvt_error(_("Corrupt volume input data: %s\n"), buffer);
 				return -1;
 			}
-			memcpy(volume->ftvt_sig, "VTBL", 4);
-			continue;
+			if (i != 0) {
+				vtbl_id = i;
+			}
+		});
+		switch (vtbl_id) {
+		case vtbl:
+			GET_SINGLE_VALUE(START, i, volume->start);
+			GET_SINGLE_VALUE(END, i, volume->end);
+			GET_SINGLE_VALUE(NUM_SEGMENTS, i, volume->num_segments);
+			GET_STRING_VALUE(DESCRIPTION, volume->ftvt_desc, 44,
+							 memset(volume->ftvt_desc + strlen(value), ' ',
+									44 - strlen(value)));
+			GET_STRING_VALUE(DATE, NULL, 1024,
+			{ if (ftvt_set_date(volumes, maxnum, value, num)) {
+				return -1;
+			} });
+			GET_SINGLE_VALUE(FLAG_VENDOR_SPECIFIC, i,
+							 volume->entry.vendor_specific);
+			GET_SINGLE_VALUE(QIC_MAJOR, i, volume->entry.ext.qic113.major);
+			GET_SINGLE_VALUE(QIC_MINOR, i, volume->entry.ext.qic113.minor);
+			GET_STRING_VALUE(VENDOR_DATA, &volume->bytes[56], 72, /**/);
+			if (volume->entry.vendor_specific &&
+				(volume->fmt_code == fmt_big ||
+				 (volume->fmt_code != fmt_big &&
+				  volume->entry.ext.qic113.major != 113))) {
+				continue; /* ignore all other tags */
+			}
+			GET_SINGLE_VALUE(FLAG_MULTI_CARTRIDGE, i,
+							 volume->entry.multi_cartridge);
+			GET_SINGLE_VALUE(FLAG_NOT_VERIFIED, i,
+							 volume->entry.not_verified);
+			GET_SINGLE_VALUE(FLAG_REDIRECTION_INHIBIT, i,
+							 volume->entry.inhibit_redirection);
+			GET_SINGLE_VALUE(FLAG_SEGMENT_SPANNING, i,
+							 volume->entry.segment_spanning);
+			GET_SINGLE_VALUE(FLAG_DIRECTORY_LAST, i,
+							 volume->entry.directory_last);
+			GET_SINGLE_VALUE(FLAG_RESERVED_6, i,
+							 volume->entry.fl_reserved_6);
+			GET_SINGLE_VALUE(FLAG_RESERVED_7, i,
+							 volume->entry.fl_reserved_7);
+			GET_SINGLE_VALUE(MULTI_CARTRIDGE_COUNT, i,
+							 volume->entry.m_no);
+			GET_SINGLE_VALUE(DIRECTORY_SIZE, i,
+							 volume->entry.dir_size);
+			GET_SINGLE_VALUE(DEVICE, i,
+							 volume->entry.device);
+			GET_SINGLE_VALUE(RESERVED_1, i,
+							 volume->entry.reserved_1);
+			GET_SINGLE_VALUE(COMPRESSION_FLAGS, i,
+							 volume->entry.cmpr);
+			GET_SINGLE_VALUE(FORMAT, i,
+							 volume->entry.format);
+			GET_SINGLE_VALUE(RESERVED_2, i,
+							 volume->entry.reserved_1);
+			GET_SINGLE_VALUE(RESERVED_3, i,
+							 volume->entry.reserved_1);
+			GET_SINGLE_VALUE(DATA_SIZE, Lu,
+							 volume->entry.data_size);
+			GET_SINGLE_VALUE(OS_VERSION, i,
+							 volume->entry.os_version);
+
+			GET_STRING_VALUE(VENDOR_EXTENSION, NULL, 1024,
+			{ char *p = value; int i; int val; int len;
+
+			if (volume->entry.ext.qic113.major == 113) {
+				len = 14;
+			} else {
+				len = 26;
+			}
+
+			for (i = 0; i < len; i ++) {
+				if (sscanf(p, "%i", &val) != 1) {
+					ftvt_error(
+						_("Corrupt volume input data: %s\n"),
+						buffer);
+					return -1;
+				}
+				if (volume->entry.ext.qic113.major == 113) {
+					volume->entry.ext.qic113.ext[i] = val;
+				} else {
+					volume->entry.ext.ext[i] = val;
+				}
+				while (isspace(*p)) p++;
+				while (*p != '\0' && !isspace(*p)) p++;
+			} });
+			GET_STRING_VALUE(PASSWORD, NULL, 1024,
+			{ char *p = value; int i; int val;
+			for (i = 0; i < sizeof(volume->entry.pwd); i ++) {
+				if (sscanf(p, "%i", &val) != 1) {
+					ftvt_error(
+						_("Corrupt volume input data: %s\n"),
+						buffer);
+					return -1;
+				}
+				volume->entry.pwd[i] = val;
+				while (isspace(*p)) p++;
+				while (*p != '\0' && !isspace(*p)) p++;
+			} });
+			GET_STRING_VALUE(SOURCE_DRIVE, volume->entry.source_drive, 16,
+							 /**/);
+			break;
+		case xtbl:
+			GET_STRING_VALUE(NAME, NULL, 1024,
+			{ char *p = value; int i; int val;
+			for (i = 0; i < 88; i ++) {
+				if (sscanf(p, "%i", &val) != 1) {
+					ftvt_error(
+						_("Corrupt volume input data: %s\n"),
+						buffer);
+					return -1;
+				}
+				volume->bytes[XTBL_NAME+i] = val;
+				while (isspace(*p)) p++;
+				while (*p != '\0' && !isspace(*p)) p++;
+			} });
+			GET_STRING_VALUE(PASSWORD, NULL, 1024,
+			{ char *p = value; int i; int val;
+			for (i = 0; i < 16; i ++) {
+				if (sscanf(p, "%i", &val) != 1) {
+					ftvt_error(
+						_("Corrupt volume input data: %s\n"),
+						buffer);
+					return -1;
+				}
+				volume->bytes[XTBL_PWD+i] = val;
+				while (isspace(*p)) p++;
+				while (*p != '\0' && !isspace(*p)) p++;
+			} });
+			GET_STRING_VALUE(RESERVED, NULL, 1024,
+			{ char *p = value; int i; int val;
+			for (i = 0; i < 20; i ++) {
+				if (sscanf(p, "%i", &val) != 1) {
+					ftvt_error(
+						_("Corrupt volume input data: %s\n"),
+						buffer);
+					return -1;
+				}
+				volume->bytes[XTBL_RES+i] = val;
+				while (isspace(*p)) p++;
+				while (*p != '\0' && !isspace(*p)) p++;
+			} });
+			break;
+		case utid:
+			GET_STRING_VALUE(NAME, NULL, 1024,
+			{ char *p = value; int i; int val;
+			for (i = 0; i < 88; i ++) {
+				if (sscanf(p, "%i", &val) != 1) {
+					ftvt_error(
+						_("Corrupt volume input data: %s\n"),
+						buffer);
+					return -1;
+				}
+				volume->bytes[UTID_NAME+i] = val;
+				while (isspace(*p)) p++;
+				while (*p != '\0' && !isspace(*p)) p++;
+			} });
+			GET_STRING_VALUE(RESERVED, NULL, 1024,
+			{ char *p = value; int i; int val;
+			for (i = 0; i < 36; i ++) {
+				if (sscanf(p, "%i", &val) != 1) {
+					ftvt_error(
+						_("Corrupt volume input data: %s\n"),
+						buffer);
+					return -1;
+				}
+				volume->bytes[UTID_RES+i] = val;
+				while (isspace(*p)) p++;
+				while (*p != '\0' && !isspace(*p)) p++;
+			} });
+			break;
+		case exvt: {
+			u_int16_t parent, child;
+			GET_SINGLE_VALUE(PARENT, i, parent);
+			memcpy(&volume->bytes[EXVT_PARENT], &parent, 2);
+			GET_SINGLE_VALUE(CHILD, i, child);
+			memcpy(&volume->bytes[EXVT_CHILD], &child, 2);
+			GET_STRING_VALUE(RESERVED, NULL, 1024,
+			{ char *p = value; int i; int val;
+			for (i = 0; i < 120; i ++) {
+				if (sscanf(p, "%i", &val) != 1) {
+					ftvt_error(
+						_("Corrupt volume input data: %s\n"),
+						buffer);
+					return -1;
+				}
+				volume->bytes[EXVT_RES+i] = val;
+				while (isspace(*p)) p++;
+				while (*p != '\0' && !isspace(*p)) p++;
+			} });
+			break;
+		}
 		}
-		GET_SINGLE_VALUE(START, i, volume->start);
-		GET_SINGLE_VALUE(END, i, volume->end);
-		GET_SINGLE_VALUE(FLAG_VENDOR_SPECIFIC, i,
-						 volume->entry.vendor_specific);
-		GET_SINGLE_VALUE(FLAG_MULTI_CARTRIDGE, i,
-						 volume->entry.multi_cartridge);
-		GET_SINGLE_VALUE(FLAG_NOT_VERIFIED, i,
-						 volume->entry.not_verified);
-		GET_SINGLE_VALUE(FLAG_REDIRECTION_INHIBIT, i,
-						 volume->entry.inhibit_redirection);
-		GET_SINGLE_VALUE(FLAG_SEGMENT_SPANNING, i,
-						 volume->entry.segment_spanning);
-		GET_SINGLE_VALUE(FLAG_DIRECTORY_LAST, i,
-						 volume->entry.directory_last);
-		GET_SINGLE_VALUE(FLAG_RESERVED_6, i,
-						 volume->entry.fl_reserved_6);
-		GET_SINGLE_VALUE(FLAG_RESERVED_7, i,
-						 volume->entry.fl_reserved_7);
-		GET_SINGLE_VALUE(MULTI_CARTRIDGE_COUNT, i,
-						 volume->entry.m_no);
-		GET_SINGLE_VALUE(DIRECTORY_SIZE, i,
-						 volume->entry.dir_size);
-		GET_SINGLE_VALUE(DEVICE, i,
-						 volume->entry.device);
-		GET_SINGLE_VALUE(RESERVED_1, i,
-						 volume->entry.reserved_1);
-		GET_SINGLE_VALUE(COMPRESSION_FLAGS, i,
-						 volume->entry.cmpr);
-		GET_SINGLE_VALUE(FORMAT, i,
-						 volume->entry.format);
-		GET_SINGLE_VALUE(RESERVED_2, i,
-						 volume->entry.reserved_1);
-		GET_SINGLE_VALUE(RESERVED_3, i,
-						 volume->entry.reserved_1);
-		GET_SINGLE_VALUE(DATA_SIZE, Lu,
-						 volume->entry.data_size);
-		GET_SINGLE_VALUE(OS_VERSION, i,
-						 volume->entry.os_version);
-		GET_STRING_VALUE(DESCRIPTION, volume->ftvt_desc, 44,
-						 memset(volume->ftvt_desc + strlen(value), ' ',
-								44 - strlen(value)));
-		GET_STRING_VALUE(DATE, NULL, 1024,
-						 { if (ftvt_set_date(volumes, maxnum, value, num)) {
-							 return -1;
-						 } });
-		GET_STRING_VALUE(VENDOR_EXTENSION, NULL, 1024,
-						 { char *p = value; int i; int val;
-						 for (i = 0; i < sizeof(volume->entry.ext)-1; i ++) {
-							 if (sscanf(p, "%i", &val) != 1) {
-								 ftvt_error(
-									 _("Corrupt volume input data: %s\n"),
-									 buffer);
-								 return -1;
-							 }
-							 volume->entry.ext[i] = val;
-							 while (isspace(*p)) p++;
-							 while (*p != '\0' && !isspace(*p)) p++;
-						 } });
-		GET_STRING_VALUE(PASSWORD, NULL, 1024,
-						 { char *p = value; int i; int val;
-						 for (i = 0; i < sizeof(volume->entry.pwd)-1; i ++) {
-							 if (sscanf(p, "%i", &val) != 1) {
-								 ftvt_error(
-									 _("Corrupt volume input data: %s\n"),
-									 buffer);
-								 return -1;
-							 }
-							 volume->entry.pwd[i] = val;
-							 while (isspace(*p)) p++;
-							 while (*p != '\0' && !isspace(*p)) p++;
-						 } });
-		GET_STRING_VALUE(SOURCE_DRIVE, volume->entry.source_drive, 16,
-						 /**/);
 	}
 	ftvt_error(_("Premature end of volume table\n"));
 	return -1;
@@ -913,7 +1336,13 @@ int ftvt_write(int tape_fd, const ftvt *volumes, u_int8_t *buffer,
 
 			memcpy(p, &volumes[i].entry, FTVT_SIZE);
 			if (volumes[i].fmt_code == fmt_big) {
-				vt->size.scsi_segs = (volumes[i].end - volumes[i].start + 1);
+				vt->size.fmt6_segs   = volumes[i].num_segments;
+				/* this will fix some pre-ftape-4.04 volumes Note that
+				 * the start and end field do not overlap with the
+				 * pre-ftape-4.04 extension field
+				 */
+				memcpy(vt->ext.qic113.start, &volumes[i].start, 4);
+				memcpy(vt->ext.qic113.end, &volumes[i].end, 4);
 			} else {
 				vt->size.se.start  = volumes[i].start;
 				vt->size.se.end    = volumes[i].end;
